import LogsData from "../components/LogsData";
import "../globals.css";

export default function Logs() {

  return (
    <div className='flex w-screen h-screen p-2 flex-col'>
      <LogsData />
    </div>
  );
}
